from abc import ABC
from fastapi.responses import JSONResponse
from starlette.exceptions import HTTPException

class NeumSinkInfo(ABC):
    def __init__(self, number_vectors_stored:int) -> None:
        self.number_vectors_stored:str = number_vectors_stored

    def as_sink_info(dct:dict):
        if dct == None:
            raise HTTPException(status_code=500, detail="[x001] An error occured on our end, please email kevin@tryneum.com to unblock you!")
        return NeumSinkInfo(
            number_vectors_stored=dct.get("number_vectors_stored", None),
        )
    
    def toJson(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['number_vectors_stored'] = self.number_vectors_stored
        return json_to_return

class NeumSinkInfoModel(JSONResponse):
    def __init__(
        self,
        content: NeumSinkInfo,
        status_code: int = 200,
    ) -> None:
        super().__init__(content, status_code, headers=None, media_type="application/json", background=None)


from abc import ABC
from starlette.exceptions import HTTPException

class CloudFile(ABC):
    def __init__(self, metadata:dict, file_identifier:str = None, id:str = None, data:str = None, type:str = None) -> None:
        self.file_identifier:str = file_identifier
        self.data:str = data
        self.metadata:dict = metadata
        self.type:str = type
        self.id:str = id

    def as_file(dct:dict):
        if dct == None:
            raise HTTPException(status_code=500, detail="[x001] An error occured on our end, please email kevin@tryneum.com to unblock you!")
        return CloudFile(
            file_identifier=dct.get("file_identifier", None),
            data=dct.get("data", None),
            metadata=dct.get("metadata", None),
            type=dct.get("type", None),
            id=dct.get("id", None)
        )
    
    def toJson(self):
        """Python does not have built in serialization. We need this logic to be able to respond in our API..

        Returns:
            _type_: the json to return
        """
        json_to_return = {}
        json_to_return['file_identifier'] = self.file_identifier
        json_to_return['data'] = self.data
        json_to_return['metadata'] = self.metadata
        json_to_return['type'] = self.type
        json_to_return['id'] = self.id
        return json_to_return
    


from abc import ABC
from neumai.Pipelines.Pipeline import Pipeline
from neumai.Pipelines.TriggerSyncTypeEnum import TriggerSyncTypeEnum
import requests
import json

class NeumClient(ABC):
    def __init__(self, api_key:str) -> None:
        self.api_key = api_key

    def createPipeline(self, pipeline:Pipeline):
        import requests
        
        url = f"https://api.neum.ai/v1/pipelines"

        # Headers
        headers = {
            "neum-api-key": self.api_key,
            "Content-Type": "application/json"
        }

        response = requests.post(url, headers=headers, json=pipeline.as_request())
        return response.json()['id']

    def getPipeline(self, pipeline_id:str):
        import requests
        import json

        url = f"https://api.neum.ai/v1/pipelines/{pipeline_id}"

        headers = {
            "accept": "application/json",
            "neum-api-key": self.api_key
        }

        response = requests.get(url, headers=headers)
        return(json.loads(response.text))
    
    def triggerPipeline(self, pipeline_id:str, sync_type:TriggerSyncTypeEnum):
        url = f"https://api.neum.ai/v1/pipelines/{pipeline_id}/trigger"

        # Headers
        headers = {
            "neum-api-key":self.api_key,
            "Content-Type": "application/json"
        }

        response = requests.post(url, headers=headers, json={ "sync_type": sync_type.value })
        return json.loads(response.text)

    def searchPipeline(self, pipeline_id:str, query:str, num_of_results:int = 3):
        url = f"https://api.neum.ai/v1/pipelines/{pipeline_id}/search"

        payload = {
            "number_of_results": num_of_results,
            "query": query
        }
        headers = {
            "accept": "application/json",
            "neum-api-key": self.api_key,
            "content-type": "application/json"
        }

        response = requests.post(url, json=payload, headers=headers)

        return json.loads(response.text)


from abc import ABC
from starlette.exceptions import HTTPException
from typing import List

class NeumVector(ABC):
    def __init__(self, id:str, vector:List[float], metadata:dict) -> None:
        self.id:str = id
        self.vector:List[float] = vector
        self.metadata:dict = metadata


from typing import List, Tuple
from neumai.EmbedConnectors.EmbedConnector import EmbedConnector
import replicate
from neumai.Shared.NeumDocument import NeumDocument

class ReplicateEmbed(EmbedConnector):
    @property
    def embed_name(self) -> str:
        return 'ReplicateEmbed'
    
    @property
    def requiredProperties(self) -> List[str]:
        return ['api_key', 'replicate_model']

    @property
    def optionalProperties(self) -> List[str]:
        return []

    def validation(self) -> bool:
        """Validate connector setup"""
        try:
            api_key = self.embed_information["api_key"]
            model = self.embed_information["replicate_model"]
        except Exception as e:
            raise ValueError("Required properties are not set.")
        return True 

    def embed(self, documents:List[NeumDocument]) -> Tuple[List, dict]:
        """Generate embeddings with Azure OpenAI"""
        api_key = self.embed_information["api_key"]
        model = self.embed_information["replicate_model"]
        client = replicate.Client(api_token=api_key)
        batch_size = 32
        batched_embeddings = []
        for i in range(0, len(documents), batch_size):
            # set end position of batch
            i_end = min(i + batch_size, len(documents))
            # get batch of texts and ids
            lines_batch = [doc.content for doc in documents[i:i_end]]
            output = client.run(
                model,
                input={"text_batch": lines_batch}
            )
            batched_embeddings  += [record for record in output['data']]
        
        info = {
            "estimated_cost":str("Not implemented"),
            "total_tokens":str("Not implemented"),
            "attempts_used":str("Not implemented")
        }

        return batched_embeddings, info

    def embed_query(self, query: str) -> List[float]:
        api_key = self.embed_information["api_key"]
        model = self.embed_information["replicate_model"]
        client = replicate.Client(api_token=api_key)
        output = client.run(
            model,
            input={"text": query}
        )

        return output['data']


from neumai.DataConnectors.DataConnector import DataConnector
from typing import List, Generator
from neumai.Shared.LocalFile import LocalFile
from neumai.Shared.CloudFile import CloudFile
from bs4 import BeautifulSoup
import tempfile
import os
import requests


class NeumWebsiteConnector(DataConnector):
    """" Neum Website Connector \n
    connector_information contains: [ url ]\n
    available metadata: [ url ]\n
    available content: [ website ]"""
  
    @property
    def connector_name(self) -> str:
        return "NeumWebsiteConnector"
    
    @property
    def requiredProperties(self) -> List[str]:
        return ["url"]

    @property
    def optionalProperties(self) -> List[str]:
        return []
    
    @property
    def availableMetadata(self) -> str:
        return ['url']

    @property
    def availableContent(self) -> str:
        return ['website']
    
    @property
    def schedule_avaialable(self) -> bool:
        return True

    @property
    def auto_sync_available(self) -> bool:
        return False
    
    @property
    def compatible_loaders(self) -> List[str]:
        return ["HTMLLoader", "AutoLoader"]
    
    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        # Send an HTTP GET request to the website
        url:str = str(self.connector_information['url'])
        clean_url = url.replace(" ", "")
        list_urls = clean_url.split(" , ")
        for u in list_urls:
            available_metadata = {'url':u}
            selected_metadata  = {k: available_metadata[k] for k in self.selector.to_metadata if k in available_metadata}
            yield CloudFile(file_identifier=u, metadata=selected_metadata, id=u)

    def connect_and_list_delta(self) -> Generator[CloudFile, None, None]:
        yield from self.connect_and_list_full()
    
    def connect_and_download(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
            response = requests.get(cloudFile.file_identifier)
            # Parse the HTML content
            soup = BeautifulSoup(response.content, 'html.parser')
            # Find the <body> element and extract its HTML content
            body = soup.find('body')
            body_html = str(body)  # Convert the body tag to a string to get its HTML content
            # Create a temporary file and write the extracted HTML to it
            with tempfile.NamedTemporaryFile(delete=False, suffix=".html", mode='w', encoding="utf-8") as temp:
                temp.write(body_html)
                yield LocalFile(file_path=temp.name, metadata=cloudFile.metadata, id=cloudFile.id)
        
    def validate(self) -> bool:
        # Check for required properties
        try:
            url:str = str(self.connector_information['url'])
        except:
            raise ValueError("Required properties not set")
        
        # Check for metadata values
        if not all(x in self.availableMetadata for x in self.selector.to_metadata):
            raise ValueError("Invalid metadata values provided")
        
        # Check to see that site exists
        try:
            response = requests.get(url)
        except Exception as e:
            raise Exception(f"Connection to website failed, check url. See Exception: {e}")      
        return True 


from datetime import datetime
from typing import List, Generator
from neumai.Shared.LocalFile import LocalFile
from neumai.Shared.CloudFile import CloudFile
from neumai.DataConnectors.DataConnector import DataConnector
import tempfile


class SharepointConnector(DataConnector):
    """" Sharepoint Connector \n
    connector_information requires:\n
    [ tenant_id, client_id, client_secret, site_id ]"""
    
    @property
    def connector_name(self) -> str:
        return "SharepointConnector"
    
    @property
    def requiredProperties(self) -> List[str]:
        return ["tenant_id", "client_id", "client_secret", "site_id"]

    @property
    def optionalProperties(self) -> List[str]:
        return []

    @property
    def availableMetadata(self) -> str:
        return ["createdDateTime", "lastModifiedDateTime", "name", "createdBy.user.email", "createdBy.user.id", "createdBy.user.displayName", "lastModifiedBy.user.email", "lastModifiedBy.user.id", "lastModifiedBy.user.displayName"]

    @property
    def availableContent(self) -> str:
        return ['file']
    
    @property
    def schedule_avaialable(self) -> bool:
        return True

    @property
    def auto_sync_available(self) -> bool:
        return False
    
    @property
    def compatible_loaders(self) -> List[str]:
        return ["AutoLoader", "HTMLLoader", "MarkdownLoader", "NeumCSVLoader", "NeumJSONLoader", "PDFLoader"]
    
    def process_folder(self, site_id:str, drive_id:str, folder_id:str, headers:dict) -> Generator[CloudFile, None, None]:
        import requests
        folder_url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/drives/{drive_id}/items/{folder_id}/children"
        folder_response = requests.get(folder_url, headers=headers)
        folder_info = folder_response.json()
        folder_items =  folder_info.get('value', [])
        for item in folder_items:
            if 'file' in item.keys():
                # It's a file, download it
                file_url = item['@microsoft.graph.downloadUrl']
                file_name = item['name']
                file_type = item['file']['mimeType']
                availableMetadata = {
                    "createdDateTime":item['createdDateTime'],
                    "lastModifiedDateTime":item['lastModifiedDateTime'],
                    "name":item['name'],
                    "createdBy.user.email":item['createdBy']['user']['email'],
                    "createdBy.user.id":item['createdBy']['user']['id'],
                    "createdBy.user.displayName":item['createdBy']['user']['displayName'],
                    "lastModifiedBy.user.email":item['createdBy']['user']['email'],
                    "lastModifiedBy.user.id":item['createdBy']['user']['id'],
                    "lastModifiedBy.user.displayName":item['createdBy']['user']['displayName'],
                }
                selected_metadata  = {k: availableMetadata[k] for k in self.selector.to_metadata if k in availableMetadata}
                yield CloudFile(file_identifier=file_url, id=file_name, type=file_type, metadata=selected_metadata, id=file_name)
            
            elif 'folder' in item.keys():
                # It's a folder, process it recursively
                yield from self.process_folder(site_id=site_id, drive_id=drive_id, folder_id=item['id'], headers=headers)
    
    def process_folder_delta(self, site_id:str, drive_id:str, folder_id:str, headers:dict, metadata_keys:List[str], last_run:datetime) -> Generator[CloudFile, None, None]:
        import requests
        folder_url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/drives/{drive_id}/items/{folder_id}/children"
        folder_response = requests.get(folder_url, headers=headers)
        folder_info = folder_response.json()
        folder_items =  folder_info.get('value', [])
        for item in folder_items:
            if 'file' in item.keys():
                # It's a file, download it
                file_url = item['@microsoft.graph.downloadUrl']
                file_name = item['name']
                file_type = item['file']['mimeType']
                availableMetadata = {
                    "createdDateTime":item['createdDateTime'],
                    "lastModifiedDateTime":item['lastModifiedDateTime'],
                    "name":item['name'],
                    "createdBy.user.email":item['createdBy']['user']['email'],
                    "createdBy.user.id":item['createdBy']['user']['id'],
                    "createdBy.user.displayName":item['createdBy']['user']['displayName'],
                    "lastModifiedBy.user.email":item['createdBy']['user']['email'],
                    "lastModifiedBy.user.id":item['createdBy']['user']['id'],
                    "lastModifiedBy.user.displayName":item['createdBy']['user']['displayName'],
                }
                if last_run < item['lastModifiedDateTime']:
                    selected_metadata  = {k: availableMetadata[k] for k in metadata_keys if k in availableMetadata}
                    yield CloudFile(file_identifier=file_url, id=file_name, type=file_type, metadata=selected_metadata, id=file_name)
            
            elif 'folder' in item.keys():
                # It's a folder, process it recursively
                yield from self.process_folder(site_id=site_id, drive_id=drive_id, folder_id=item['id'], headers=headers)

    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        """Connect to source and download file into local storage"""
        import requests

        #Required
        tenant_id = self.connector_information['tenant_id']
        client_id = self.connector_information['client_id']
        client_secret = self.connector_information['client_secret']
        site_id = self.connector_information['site_id']

        #Auth
        token_url = f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token"
        token_data = {
            'grant_type': 'client_credentials',
            'client_id': client_id,
            'client_secret': client_secret,
            'scope': 'https://graph.microsoft.com/.default'
        }
        token_r = requests.post(token_url, headers={"Content-Type":"application/x-www-form-urlencoded"}, data=token_data)
        token = token_r.json().get('access_token')

        # Get all document libraries in the site
        headers = {
            'Authorization': f'Bearer {token}',
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }
        doc_libraries_url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/drives"
        doc_libraries_response = requests.get(doc_libraries_url, headers=headers)
        doc_libraries_info = doc_libraries_response.json()

        # Iterate through all document libraries
        for doc_library in doc_libraries_info.get('value', []):
            doc_library_id = doc_library['id']
            yield from self.process_folder(site_id=site_id, drive_id=doc_library_id, folder_id='root', headers=headers)

    def connect_and_list_delta(self, last_run:datetime) -> Generator[LocalFile, None, None]:
        """Check for changes in the source"""
        """Code to be pushed to a worker and run on a schedule"""
        import requests

        #Required
        tenant_id = self.connector_information['tenant_id']
        client_id = self.connector_information['client_id']
        client_secret = self.connector_information['client_secret']
        site_id = self.connector_information['site_id']

        #Auth
        token_url = f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token"
        token_data = {
            'grant_type': 'client_credentials',
            'client_id': client_id,
            'client_secret': client_secret,
            'scope': 'https://graph.microsoft.com/.default'
        }
        token_r = requests.post(token_url, headers={"Content-Type":"application/x-www-form-urlencoded"}, data=token_data)
        token = token_r.json().get('access_token')

        # Get all document libraries in the site
        headers = {
            'Authorization': f'Bearer {token}',
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }
        doc_libraries_url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/drives"
        doc_libraries_response = requests.get(doc_libraries_url, headers=headers)
        doc_libraries_info = doc_libraries_response.json()

        # Iterate through all document libraries
        for doc_library in doc_libraries_info.get('value', []):
            doc_library_id = doc_library['id']
            yield from self.process_folder_delta(site_id=site_id, drive_id=doc_library_id, folder_id='root', headers=headers, last_run=last_run)
        
    def connect_and_download(self, cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
        """Connect to source and download file into local storage"""
        #Download file
        id = cloudFile.id
        file_identifier = cloudFile.file_identifier
        type = cloudFile.type
        metadata = cloudFile.metadata
        import requests
        file_r = requests.get(file_identifier)
        with tempfile.NamedTemporaryFile(delete=False, suffix=f"_{id}") as temp_file:
            temp_file.write(file_r.content)
        yield LocalFile(file_path=temp_file.name, metadata=metadata, id=id, type=type)
    
    def validate(self) -> bool:
        try:
            tenant_id = self.connector_information['tenant_id']
            client_id = self.connector_information['client_id']
            client_secret = self.connector_information['client_secret']
            site_id = self.connector_information['site_id']
        except:
            raise ValueError("Required properties not set")
        
        if not all(x in self.availableMetadata for x in self.selector.to_metadata):
            raise ValueError("Invalid metadata values provided")
        
        try:
            import requests
            #Auth
            token_url = f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token"
            token_data = {
                'grant_type': 'client_credentials',
                'client_id': client_id,
                'client_secret': client_secret,
                'scope': 'https://graph.microsoft.com/.default'
            }
            token_r = requests.post(token_url, headers={"Content-Type":"application/x-www-form-urlencoded"}, data=token_data)
            token = token_r.json().get('access_token')

            # Get all document libraries in the site
            headers = {
                'Authorization': f'Bearer {token}',
                'Accept': 'application/json',
                'Content-Type': 'application/json'
            }
            doc_libraries_url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/drives"
            doc_libraries_response = requests.get(doc_libraries_url, headers=headers)
            doc_libraries_info = doc_libraries_response.json()
        except Exception as e:
            raise Exception(f"Connection to Sharepoint failed, check credentials. See Exception: {e}")       
        return True 


from typing import List, Tuple
from neumai.EmbedConnectors.EmbedConnector import EmbedConnector
from langchain.embeddings.openai import OpenAIEmbeddings
from neumai.Shared.NeumDocument import NeumDocument

class OpenAIEmbed(EmbedConnector):
    """" OpenAI Embed Connector \n
    embed_information required: [ api_key ]"""

    @property
    def embed_name(self) -> str:
        return 'OpenAIEmbed'
    
    @property
    def requiredProperties(self) -> List[str]:
        return ["api_key"]

    @property
    def optionalProperties(self) -> List[str]:
        return ['max_retries', 'chunk_size']

    def validation(self) -> bool:
        """Validate connector setup"""
        try:
            api_key = self.embed_information['api_key']
        except:
            raise ValueError("Required properties not set")
        try:
            organization = self.embed_information.get('organization', None)
            OpenAIEmbeddings(max_retries=20, openai_api_key=api_key, openai_organization=organization, chunk_size=1000)
        except Exception as e:
            raise ValueError(f"OpenAI couldn't be initialized. See exception: {e}")
        return True 

    def embed(self, documents:List[NeumDocument]) -> Tuple[List, dict]:
        """Generate embeddings with OpenAI"""
        max_retries = self.embed_information.get('max_retries', 20)
        chunk_size = self.embed_information.get('chunk_size', 1000)
        organization = self.embed_information.get('organization', None)

        embedding = OpenAIEmbeddings(max_retries=max_retries, openai_api_key=self.embed_information['api_key'], openai_organization=organization, chunk_size=chunk_size)
        embeddings = []
        texts = [x.content for x in documents]
        # do we want to persist some embeddings if they were able to be wrriten but not another "batch" of them? or should we treat all texts as an atomic operation
        embeddings  = embedding.embed_documents(texts=texts)
        #cost_per_token = 0.000000001 # ADA-002 as of Sept 2023
        info = {
            "estimated_cost":str("Not implemented"),
            "total_tokens":str("Not implemented"),
            "attempts_used":str("Not implemented")
        }
        return embeddings,info

    def embed_query(self, query: str) -> List[float]:
        import openai
        openai.api_key = self.embed_information["api_key"]
        openai.organization = self.embed_information.get("organization", "")
        result = openai.Embedding.create(input=query, model="text-embedding-ada-002")
        return result['data'][0]['embedding']



from datetime import datetime
from neumai.DataConnectors.DataConnector import DataConnector
from typing import List, Generator
from azure.storage.blob import BlobClient, ContainerClient
from neumai.Shared.LocalFile import LocalFile
from neumai.Shared.CloudFile import CloudFile
import tempfile
import os


class AzureBlobConnector(DataConnector):
    """" Neum File Connector """

    @property
    def connector_name(self) -> str:
        return "AzureBlobConnector"
    
    @property
    def requiredProperties(self) -> List[str]:
        return ["connection_string", "container_name"]

    @property
    def optionalProperties(self) -> List[str]:
        return []
    
    @property
    def availableMetadata(self) -> str:
        return ['name', 'last_modified', 'creation_time', 'last_access_on']

    @property
    def availableContent(self) -> str:
        return ['file']
    
    @property
    def schedule_avaialable(self) -> bool:
        return True

    @property
    def auto_sync_available(self) -> bool:
        return False
    
    @property
    def compatible_loaders(self) -> List[str]:
        return ["AutoLoader", "HTMLLoader", "MarkdownLoader", "NeumCSVLoader", "NeumJSONLoader", "PDFLoader"]
    
    def connect_and_list_full(self) -> Generator[CloudFile, None, None]:
        # Connect to Azure Blob Storage
        connection_string = self.connector_information['connection_string']
        container_name = self.connector_information['container_name']
        container = ContainerClient.from_connection_string(
            conn_str=connection_string, container_name=container_name
        )

        #Process files
        file_list = container.list_blobs()
        for file in file_list:
            name = file.name
            metadata = {
                "creation_time": file.creation_time,
                "last_modified": file.last_modified,
                "last_access_on": file.last_accessed_on,
            }
            selected_metadata  = {k: metadata[k] for k in self.selector.to_metadata if k in metadata}
            yield CloudFile(file_identifier=name, metadata=selected_metadata, id = name)

    def connect_and_list_delta(self, last_run:datetime) -> Generator[CloudFile, None, None]:
        # Connect to Azure Blob Storage
        connection_string = self.connector_information['connection_string']
        container_name = self.connector_information['container_name']
        container = ContainerClient.from_connection_string(
            conn_str=connection_string, container_name=container_name
        )

        #Process files
        file_list = container.list_blobs()
        for file in file_list:
            last_update_date = file.last_modified
            if(last_run < last_update_date):
                name = file.name
                metadata = {
                    "creation_time": file.creation_time,
                    "last_modified": file.last_modified,
                    "last_access_on": file.last_accessed_on,
                }
                selected_metadata  = {k: metadata[k] for k in self.selector.to_metadata if k in metadata}
                yield CloudFile(file_identifier=name, metadata=selected_metadata, id=name)

    
    def connect_and_download(self,  cloudFile:CloudFile) -> Generator[LocalFile, None, None]:
        # Connect to Azure Blob Storage
        connection_string = self.connector_information['connection_string']
        container_name = self.connector_information['container_name']
        
        client = BlobClient.from_connection_string(conn_str=connection_string, container_name=container_name, blob_name=cloudFile.file_identifier)
        with tempfile.TemporaryDirectory() as temp_dir:
            file_path = f"{temp_dir}/{container_name}/{cloudFile.file_identifier}"
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            with open(f"{file_path}", "wb") as file:
                blob_data = client.download_blob()
                blob_data.readinto(file)
            yield LocalFile(file_path=file_path, metadata=cloudFile.metadata, id=cloudFile.id)
        
    def validate(self) -> bool:
        try:
            connection_string = self.connector_information['connection_string']
            container_name = self.connector_information['container_name']
        except:
            raise ValueError("Required properties not set")
        
        if not all(x in self.availableMetadata for x in self.selector.to_metadata):
            raise ValueError("Invalid metadata values provided")
        
        try:
            container = ContainerClient.from_connection_string(
                conn_str=connection_string, container_name=container_name
            )
        except Exception as e:
            raise Exception(f"Connection to Azure Blob Storage failed, check credentials. See Exception: {e}")   
        return True     


    
 


from typing import List, Optional, Tuple
from neumai.SinkConnectors.SinkConnector import SinkConnector
from neumai.Shared.NeumVector  import NeumVector
from neumai.Shared.NeumSinkInfo import NeumSinkInfo
from neumai.Shared.NeumSearch import NeumSearchResult

class WeaviateSink(SinkConnector):
    """ Weaviate Sink\n
    sink_information requires : [ url , api_key ]"""

    @property
    def sink_name(self) -> str:
        return 'WeaviateSink'
    
    @property
    def requiredProperties(self) -> List[str]:
        return ['url', 'api_key']

    @property
    def optionalProperties(self) -> List[str]:
        return ['class_name', 'num_workers', 'shard_count', 'batch_size', 'is_dynamic_batch', 'batch_connection_error_retries']

    def validation(self) -> bool:
        """Validate connector setup"""
        import weaviate
        try:
            url = self.sink_information["url"]
            if 'https' not in url:
                api_key = self.sink_information["api_key"]
        except:
            raise ValueError("Required properties not set")
        try:
            if 'https' not in url:
                client = weaviate.Client(
                    url=url
                )
            else:
                api_key = self.sink_information["api_key"]
                client = weaviate.Client(
                    url=url,
                    auth_client_secret=weaviate.AuthApiKey(api_key=api_key),
                )
        except Exception as e:
            raise ValueError(f"Weaviate couldn't be initialized. See exception: {e}")
        return True 

    def store(self, pipeline_id: str, vectors_to_store:List[NeumVector], task_id:str = "") -> Tuple[List, dict]:
        import weaviate
        from weaviate.util import generate_uuid5
        url = self.sink_information["url"]
        num_workers = self.sink_information.get('num_workers', 1)
        shard_count = self.sink_information.get('shard_count', 1)
        batch_size = self.sink_information.get('batch_size', 100)
        is_dynamic_batch = self.sink_information.get('is_dynamic_batch', False)
        batch_connection_error_retries = self.sink_information.get('batch_connection_error_retries', 3)
        class_name = self.sink_information.get('class_name', f"pipeline_{pipeline_id.replace('-','_')}")
        partial_failure = {'did_fail': False, 'latest_failure': None, 'number_of_failures': 0}

        if 'https' not in url:
            client = weaviate.Client(
                url=url
            )
        else:
            api_key = self.sink_information["api_key"]
            client = weaviate.Client(
                url=url,
                auth_client_secret=weaviate.AuthApiKey(api_key=api_key),
            )
        try:
            client.schema.create_class({
                "class": class_name,
                "shardingConfig":{"desiredCount":shard_count},
            })
        except weaviate.UnexpectedStatusCodeException as e:
            if 'already exists' not in e.message: # We got an error that is nto because the class already exists. can we check the class before maybe?. In this case, we should throw
                print(f"Error when creating class in weaviate.. Skipping task id {task_id}: {e}")
                raise e
            
        with client.batch.configure(
            batch_size=batch_size,
            callback=lambda results: self.check_batch_result(results, task_id, partial_failure),
            num_workers=num_workers,
            dynamic=is_dynamic_batch,
            connection_error_retries=batch_connection_error_retries
        ) as batch:
            for i in range(0, len(vectors_to_store)):
                try:
                    batch.add_data_object(
                        data_object=vectors_to_store[i].metadata,
                        class_name=class_name,
                        vector=vectors_to_store[i].vector,
                        uuid=generate_uuid5(vectors_to_store[i].id)
                    )
                except Exception as e:
                    print(f"[ERROR] Got exception from Weaviate when adding data object to batch for task {task_id} and batch # {i}. Exception: {str(e)}")
                    raise e
        # We can define the logic later here as to what constitutes a failure
        if partial_failure['number_of_failures'] > 5:
            raise Exception(f"Insertion to weaviate failed - Received more than 5 number of failures when batching. Latest error when batching was: {partial_failure['latest_failure']}")
        return len(vectors_to_store)#, partial_failure

    def check_batch_result(self, results: Optional[List[dict[str, any]]], task_id: str, partial_failure: dict):
        if results is not None:
            for result in results:
                if "result" in result and "errors" in result["result"]:
                    if "error" in result["result"]["errors"]:
                        print(f"[ERROR] Task Id {task_id} encountered an error when batching to weaviate {result['result']}")
                        partial_failure['did_fail'] = True
                        partial_failure['latest_failure'] = result["result"]["errors"]["error"]
                        partial_failure['number_of_failures'] += 1

    def search(self, vector: List[float], number_of_results: int, pipeline_id: str) -> List[NeumSearchResult]:
        import weaviate
        from weaviate.util import _capitalize_first_letter
        api_key = self.sink_information["api_key"]
        url = self.sink_information['url']
        # Weaviate requires first letter to be capitalized
        class_name = _capitalize_first_letter(self.sink_information.get('class_name', f"Pipeline_{pipeline_id.replace('-', '_')}"))
        client = weaviate.Client(
            url=url,
            auth_client_secret=weaviate.AuthApiKey(api_key=api_key),
        )

        try:
            class_schema = client.schema.get(class_name)
        except Exception as e:
            raise Exception(f"There was an error retrieving the class schema from weaviate")

        full_class_schema_properties = [property['name'] for property in class_schema['properties']]
        matches = []
        try:
            search_result = (
                client.query
                .get(class_name=class_name, properties=full_class_schema_properties)
                .with_near_vector(content = {
                    'vector' : vector
                })
                .with_limit(number_of_results)
                .with_additional(['id','certainty'])
                .do()
            )

            for result in search_result["data"]["Get"][class_name]:
                # unify our api with the metadata.. or just return whatever metadata we have. (?)
                matches.append(NeumSearchResult(id=result['_additional']['id'], score=result['_additional']['certainty'], metadata= {k: v for k, v in result.items() if k != "_additional"}))
        except Exception as e:
            raise Exception(f"There was an error querying weaviate")
        return matches

    def info(self, pipeline_id:str) -> NeumSinkInfo:
        import weaviate
        from weaviate.util import _capitalize_first_letter
        api_key = self.sink_information["api_key"]
        url = self.sink_information['url']
        
        class_name = _capitalize_first_letter(self.sink_information.get('class_name', f"Pipeline_{pipeline_id.replace('-', '_')}"))
        client = weaviate.Client(
            url=url,
            auth_client_secret=weaviate.AuthApiKey(api_key=api_key),
        )
        try:
            response = (
                client.query
                .aggregate(class_name=class_name)
                .with_meta_count()
                .do()
            )
            vectors_stored_in_class = response["data"]["Aggregate"][class_name]["meta"]["count"]
            return NeumSinkInfo(number_vectors_stored=vectors_stored_in_class)
        except Exception as e:
            raise Exception(f"There was an error querying weaviate")


from typing import List, Tuple
from neumai.EmbedConnectors.EmbedConnector import EmbedConnector
from neumai.Shared.NeumDocument import NeumDocument

class AzureOpenAIEmbed(EmbedConnector):
    @property
    def embed_name(self) -> str:
        return 'AzureOpenAIEmbed'
    
    @property
    def requiredProperties(self) -> List[str]:
        return ['deployment_name', 'api_key', 'endpoint']

    @property
    def optionalProperties(self) -> List[str]:
        return []

    def validation(self) -> bool:
        """Validate connector setup"""
        try:
            deployment_name = self.embed_information["deployment_name"]
            api_key = self.embed_information["api_key"]
            api_base = self.embed_information["endpoint"]
        except Exception as e:
            raise ValueError("Required properties are not set.")
        return True 

    def embed(self, documents:List[NeumDocument]) -> Tuple[List, dict]:
        """Generate embeddings with Azure OpenAI"""
        import openai
        cost_per_token = 0.000000001 # ADA-002 as of Sept 2023
        batch_size = 32 # how does changing this affect ?
        # Should we generate batched embeddings, store, then batch embeddings, store. or should we gneralte all embeddings then store? what's perf diff?
        batched_embeddings = []
        deployment_name = self.embed_information["deployment_name"]
        openai.api_key = self.embed_information["api_key"]
        openai.api_base = self.embed_information["endpoint"]
        openai.api_type = "azure"
        openai.api_version = "2023-05-15"
        for i in range(0, len(documents), batch_size):
            # set end position of batch
            i_end = min(i + batch_size, len(documents))
            # get batch of texts and ids
            lines_batch = [doc.content for doc in documents[i:i_end]]
            result = openai.Embedding.create(input=lines_batch, engine=deployment_name)
            batched_embeddings  += [record['embedding'] for record in result['data']]
        
        info = {
            "estimated_cost":str("Not implemented"),
            "total_tokens":str("Not implemented"),
            "attempts_used":str("Not implemented")
        }

        return batched_embeddings, info
    
    def embed_query(self, query: str) -> List[float]:
        import openai
        openai.api_type = "azure"
        openai.api_key = self.embed_information["api_key"]
        openai.api_base = self.embed_information["endpoint"]
        openai.api_version = "2023-05-15"
        deployment_name = self.embed_information["deployment_name"]
        result = openai.Embedding.create(input=query, model="text-embedding-ada-002")
        return result['data'][0]['embedding']


from neumai.SinkConnectors import (
    PineconeSink,
    WeaviateSink,
    QdrantSink,
    SingleStoreSink,
    SupabaseSink
)
from starlette.exceptions import HTTPException

def as_sink(dct:dict):
    if dct == None:
        raise HTTPException(status_code=500, detail="[x001] An error occured on our end, please email kevin@tryneum.com to unblock you!")
    sink_name = dct.get("sink_name", None)
    sink_information = dct.get("sink_information", None)
    if sink_name == "PineconeSink":
        return PineconeSink(sink_information=sink_information)
    elif sink_name == "WeaviateSink":
        return WeaviateSink(sink_information=sink_information)
    elif sink_name == "QdrantSink":
        return QdrantSink(sink_information=sink_information)
    elif sink_name == "SingleStoreSink":
        return SingleStoreSink(sink_information=sink_information)
    elif sink_name == "SupabaseSink":
        return SupabaseSink(sink_information=sink_information)
    return None


from .PineconeSink import PineconeSink
from .QdrantSink import QdrantSink
from .WeaviateSink import WeaviateSink
from .SingleStoreSink import SingleStoreSink
from .SupabaseSink import SupabaseSink
from .SinkHelper import as_sink
from .SinkConnector import SinkConnector
